<?php declare(strict_types=1);

namespace Sharkk\Router;

class SegmentRouter implements RouterInterface
{
	private array $routes = [];

	/**
	 * Add a route to the route tree. The route must be a URI path, and contain dynamic segments
	 * using a colon prefix. (:id, :slug, etc)
	 *
	 * Example:
	 * `$r->add('GET', '/posts/:id', function($id) { echo "Viewing post $id"; });`
	 */
	public function add(string $method, string $route, callable $handler): RouterInterface
	{
		// Recursively build the node tree
		$this->addNode(
			$this->routes,
			array_map( // Expand the route into segments and make dynamic segments into a common placeholder
				fn($segment) => str_starts_with($segment, ':') ? ':x' : $segment,
				explode('/', trim($route, '/'))
			),
			$method,
			$handler
		);

		// Add the handler to the last node
		$node[$method] = $handler;

		return $this;
	}

	private function addNode(array &$node, array $segments, string $method, callable $handler): void
    {
		// Base case: if no more segments, add the handler
		if (empty($segments)) {
			$node[$method] = $handler;
			return;
		}

		// Get the current segment
		$segment = array_shift($segments);

		// Skip empty segments
		if ($segment === '') {
			$this->addNode($node, $segments, $method, $handler);
			return;
		}

		// Ensure the segment exists in the node
		if (!isset($node[$segment])) {
			$node[$segment] = [];
		}

		// Recur for the next segment
		$this->addNode($node[$segment], $segments, $method, $handler);
    }

	/**
	 * Perform a lookup in the route tree for a given method and URI. Returns an array with a result code,
	 * a handler if found, and any dynamic parameters. Codes are 200 for success, 404 for not found, and
	 * 405 for method not allowed.
	 *
	 * @return array ['code', 'handler', 'params']
	 */
	public function lookup(string $method, string $uri): array
	{
		// node is a reference to our current location in the node tree
		$node = $this->routes;

		// if the URI is just a slash, we can return the handler for the root node
		if ($uri === '/') {
			return isset($node[$method])
				? ['code' => 200, 'handler' => $node[$method], 'params' => null]
				: ['code' => 405, 'handler' => null, 'params' => null];
		}

		// params will hold any dynamic segments we find
		$params = array_reduce(
			explode('/', trim($uri, '/')),
			function ($carry, $segment) use (&$node) {
				if (isset($node[$segment])) {
					$node = $node[$segment];
				} elseif (isset($node[':x'])) {
					$carry[] = $segment;
					$node = $node[':x'];
				} else {
					throw new \Exception('404');
				}
				return $carry;
			},
			[]
		);

		// if we found a handler for the method, return it and any params. if not, return a 405
		return isset($node[$method])
			? [
				'code' => 200,
				'handler' => $node[$method],
				'params' => $params ?? []]
			: [
				'code' => 405,
				'handler' => null,
				'params' => []
			];
	}

	/**
	 * Clear all routes from the router.
	 */
	public function clear(): RouterInterface
	{
		$this->routes = [];
		return $this;
	}

	/**
	 * Dump the route tree as an array.
	 */
	public function dump(): array
	{
		return $this->routes;
	}

	/**
	 * Register a GET route.
	 */
	public function get(string $route, callable $handler): RouterInterface
	{
		return $this->add('GET', $route, $handler);
	}

	/**
	 * Register a POST route.
	 */
	public function post(string $route, callable $handler): RouterInterface
	{
		return $this->add('POST', $route, $handler);
	}

	/**
	 * Register a PUT route.
	 */
	public function put(string $route, callable $handler): RouterInterface
	{
		return $this->add('PUT', $route, $handler);
	}

	/**
	 * Register a PATCH route.
	 */
	public function patch(string $route, callable $handler): RouterInterface
	{
		return $this->add('PATCH', $route, $handler);
	}

	/**
	 * Register a DELETE route.
	 */
	public function delete(string $route, callable $handler): RouterInterface
	{
		return $this->add('DELETE', $route, $handler);
	}

	/**
	 * Register a HEAD route.
	 */
	public function head(string $route, callable $handler): RouterInterface
	{
		return $this->add('HEAD', $route, $handler);
	}
}
